package cz.muni.fi.pv168.project.db.subtask;

import cz.muni.fi.pv168.project.data.task.SubTask;
import cz.muni.fi.pv168.project.db.DataAccessException;
import cz.muni.fi.pv168.project.db.interfaces.OneToNDao;

import javax.sql.DataSource;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.List;

public class SubTaskDao implements OneToNDao<SubTask> {

    private final DataSource dataSource;

    public SubTaskDao(DataSource dataSource) {
        this.dataSource = dataSource;
    }

    @Override
    public List<SubTask> getAllEntitiesFor(Long taskId) {
        if (taskId == null) {
            throw new IllegalArgumentException("Cannot look for subtasks of null ID parent");
        }

        try (var connection = dataSource.getConnection();
             var st = connection.prepareStatement("SELECT ID, TITLE, IS_COMPLETED FROM SUBTASK WHERE TASK_ID = ?")) {
            st.setLong(1, taskId);

            List<SubTask> subTasks = new ArrayList<>();
            try (var rs = st.executeQuery()) {
                while (rs.next()) {
                    subTasks.add(getFullSubTaskTaskFromDb(rs));
                }

                return subTasks;
            }
        } catch (SQLException ex) {
            throw new DataAccessException("Failed to load subtasks with task id " + taskId, ex);
        }
    }

    @Override
    public void deleteAllEntitiesFor(Long taskId) {
        var subTasks = getAllEntitiesFor(taskId);
        for (var subTask : subTasks) {
            delete(subTask);
        }
    }

    @Override
    public void addEntityFor(Long taskId, SubTask entity) {
        if (entity.getId() != null) {
            throw new IllegalArgumentException("Subtask already has ID: " + entity);
        }

        try (var connection = dataSource.getConnection();
             var st = connection.prepareStatement(
                     "INSERT INTO SUBTASK (TASK_ID, TITLE, IS_COMPLETED) VALUES (?, ?, ?)",
                     Statement.RETURN_GENERATED_KEYS)) {
            st.setLong(1, taskId);
            st.setString(2, entity.getTitle());
            st.setBoolean(3, entity.isCompleted());

            st.executeUpdate();

            try (var rs = st.getGeneratedKeys()) {
                if (rs.getMetaData().getColumnCount() != 1) {
                    throw new DataAccessException("Failed to fetch generated key: compound key returned for subtask: " + entity);
                }
                if (rs.next()) {
                    entity.setId(rs.getLong(1));
                } else {
                    throw new DataAccessException("Failed to fetch generated key: no key returned for subtask: " + entity);
                }
                if (rs.next()) {
                    throw new DataAccessException("Failed to fetch generated key: multiple keys returned for subtask: " + entity);
                }
            }
        } catch (SQLException ex) {
            throw new DataAccessException("Failed to store subtask " + entity, ex);
        }
    }

    @Override
    public void update(SubTask entity) {
        if (entity.getId() == null) {
            throw new IllegalArgumentException("Subtask has null ID: " + entity);
        }

        try (var connection = dataSource.getConnection();
             var st = connection.prepareStatement(
                     "UPDATE SUBTASK SET TITLE = ?, IS_COMPLETED = ? WHERE ID = ?")) {
            st.setString(1, entity.getTitle());
            st.setBoolean(2, entity.isCompleted());
            st.setLong(3, entity.getId());

            int rowsUpdated = st.executeUpdate();
            if (rowsUpdated == 0) {
                throw new DataAccessException("Failed to update non-existing subtask: " + entity);
            }
        } catch (SQLException ex) {
            throw new DataAccessException("Failed to update subtask " + entity, ex);
        }
    }

    @Override
    public void delete(SubTask entity) {
        if (entity.getId() == null) {
            throw new IllegalArgumentException("Subtask has null ID: " + entity);
        }

        try (var connection = dataSource.getConnection();
             var st = connection.prepareStatement("DELETE FROM SUBTASK WHERE ID = ?")) {
            st.setLong(1, entity.getId());

            int rowsDeleted = st.executeUpdate();
            if (rowsDeleted == 0) {
                throw new DataAccessException("Failed to delete non-existing subtask: " + entity);
            }
        } catch (SQLException ex) {
            throw new DataAccessException("Failed to delete subtask " + entity, ex);
        }
    }

    @Override
    public SubTask fetch(Long id) {
        if (id == null) {
            throw new IllegalArgumentException("Cannot look for subtask of null ID");
        }

        try (var connection = dataSource.getConnection();
             var st = connection.prepareStatement("SELECT ID, TITLE, IS_COMPLETED FROM SUBTASK WHERE ID = ?")) {
            st.setLong(1, id);

            try (var rs = st.executeQuery()) {
                if (rs.next()) {
                    var task = getFullSubTaskTaskFromDb(rs);
                    if (rs.next()) {
                        throw new DataAccessException("Multiple subtasks with id " + id + " found");
                    }
                    return task;
                }
                return null;
            }
        } catch (SQLException ex) {
            throw new DataAccessException("Failed to load subtask with id " + id, ex);
        }
    }

    private SubTask getFullSubTaskTaskFromDb(ResultSet rs) throws SQLException {
        var subTask = new SubTask(
            rs.getString("TITLE"),
            rs.getBoolean("IS_COMPLETED")
        );
        subTask.setId(rs.getLong("ID"));
        return subTask;
    }
}
